@@ -73,7 +73,7 @@ int checkaccess(const char *path, const char *username, int mode)
 	struct passwd *user;
 	int ret = 0, status;
 
-    if(!getuid()) {
+    if(!geteuid()) {
 
 	if((user = getpwnam(username)) == NULL) {
 	    return -1;
@@ -113,28 +113,6 @@ int checkaccess(const char *path, const char *username, int mode)
     return ret;
 }
 
-int filecopy(const char *src, const char *dest)
-{
-	char buffer[FILEBUFF];
-	int s, d, bytes;
-
-    if((s = open(src, O_RDONLY)) == -1)
-	return -1;
-
-    if((d = open(dest, O_CREAT|O_WRONLY|O_TRUNC)) == -1) {
-	close(s);
-	return -1;
-    }
-
-    while((bytes = read(s, buffer, FILEBUFF)) > 0)
-	write(d, buffer, bytes);
-
-    close(s);
-
-    /* njh@bandsman.co.uk: check result of close for NFS file */
-    return close(d);
-}
-
 int isnumb(const char *str)
 {
 	int i;
