@@ -1,5 +1,5 @@
 /*
- *  Copyright (C) 2002 - 2004 Tomasz Kojm <tkojm@clamav.net>
+ *  Copyright (C) 2002 - 2005 Tomasz Kojm <tkojm@clamav.net>
  *  HTTP/1.1 compliance by Arkadiusz Miskiewicz <misiek@pld.org.pl>
  *  Proxy support by Nigel Horne <njh@bandsman.co.uk>
  *  Proxy authorization support by Gernot Tenchio <g.tenchio@telco-tech.de>
@@ -44,14 +44,21 @@
 #include "manager.h"
 #include "notify.h"
 #include "memory.h"
+#include "output.h"
+#include "../libclamav/others.h"
+#include "../libclamav/str.h" /* cli_strtok */
+#include "dns.h"
+
 
 int downloadmanager(const struct cfgstruct *copt, const struct optstruct *opt, const char *hostname)
 {
 	time_t currtime;
-	int ret, updated = 0, signo = 0;
-	char ipaddr[16];
+	int ret, updated = 0, signo = 0, ttl = -1;
+	char ipaddr[16], *dnsreply = NULL, *pt;
 	struct cfgstruct *cpt;
-
+#ifdef HAVE_RESOLV_H
+	const char *dnsdbinfo;
+#endif
 
     time(&currtime);
     mprintf("ClamAV update process started at %s", ctime(&currtime));
@@ -59,42 +66,113 @@ int downloadmanager(const struct cfgstruct *copt, const struct optstruct *opt, c
 
 #ifndef HAVE_GMP
     mprintf("SECURITY WARNING: NO SUPPORT FOR DIGITAL SIGNATURES\n");
+    mprintf("See the FAQ at http://www.clamav.net/faq.html for an explanation.\n");
     logg("SECURITY WARNING: NO SUPPORT FOR DIGITAL SIGNATURES\n");
+    logg("See the FAQ at http://www.clamav.net/faq.html for an explanation.\n");
 #endif
 
+#ifdef HAVE_RESOLV_H
+    if((cpt = cfgopt(copt, "DNSDatabaseInfo")))
+	dnsdbinfo = cpt->strarg;
+    else
+	dnsdbinfo = "current.cvd.clamav.net";
+
+    if(optl(opt, "no-dns")) {
+	dnsreply = NULL;
+    } else {
+	if((dnsreply = txtquery(dnsdbinfo, &ttl))) {
+	    mprintf("*TTL: %d\n", ttl);
+
+	    if((pt = cli_strtok(dnsreply, 3, ":"))) {
+		    int rt;
+		    time_t ct;
+
+		rt = atoi(pt);
+		free(pt);
+		time(&ct);
+		if((int) ct - rt > 10800) {
+		    mprintf("WARNING: DNS record is older than 3 hours.\n");
+		    logg("WARNING: DNS record is older than 3 hours.\n");
+		    free(dnsreply);
+		    dnsreply = NULL;
+		}
+
+	    } else {
+		free(dnsreply);
+		dnsreply = NULL;
+	    }
+
+	    if(dnsreply && (pt = cli_strtok(dnsreply, 0, ":"))) {
+		mprintf("*Software version from DNS: %s\n", pt);
+		if(!strstr(cl_retver(), "devel") && !strstr(cl_retver(), "rc")) {
+		    if(strcmp(cl_retver(), pt)) {
+			mprintf("WARNING: Your ClamAV installation is OUTDATED - please update immediately!\n");
+			mprintf("WARNING: Local version: %s Recommended version: %s\n", cl_retver(), pt);
+			logg("WARNING: Your ClamAV installation is OUTDATED - please update immediately!\n");
+			logg("WARNING: Local version: %s Recommended version: %s\n", cl_retver(), pt);
+		    }
+		}
+		free(pt);
+
+	    } else {
+		if(dnsreply) {
+		    free(dnsreply);
+		    dnsreply = NULL;
+		}
+	    }
+	}
+
+	if(!dnsreply) {
+	    mprintf("WARNING: Invalid DNS reply. Falling back to HTTP mode.\n");
+	    logg("WARNING: Invalid DNS reply. Falling back to HTTP mode.\n");
+	}
+    }
+#endif /* HAVE_RESOLV_H */
+
     memset(ipaddr, 0, sizeof(ipaddr));
 
-    if((ret = downloaddb(DB1NAME, "main.cvd", hostname, ipaddr, &signo, copt)) > 50)
+    if((ret = downloaddb(DB1NAME, "main.cvd", hostname, ipaddr, &signo, copt, dnsreply)) > 50) {
+	if(dnsreply)
+	    free(dnsreply);
+
 	return ret;
-    else if(ret == 0)
+
+    } else if(ret == 0)
 	updated = 1;
 
     /* if ipaddr[0] != 0 it will use it to connect to the web host */
-    if((ret = downloaddb(DB2NAME, "daily.cvd", hostname, ipaddr, &signo, copt)) > 50)
+    if((ret = downloaddb(DB2NAME, "daily.cvd", hostname, ipaddr, &signo, copt, dnsreply)) > 50) {
+	if(dnsreply)
+	    free(dnsreply);
+
 	return ret;
-    else if(ret == 0)
+
+    } else if(ret == 0)
 	updated = 1;
 
+    if(dnsreply)
+	free(dnsreply);
+
     if(updated) {
 	if(cfgopt(copt, "HTTPProxyServer")) {
-	    mprintf("Database updated (%d signatures) from %s.\n", signo, hostname);
-	    logg("Database updated (%d signatures) from %s.\n", signo, hostname);
+	    mprintf("Database updated (%d signatures) from %s\n", signo, hostname);
+	    logg("Database updated (%d signatures) from %s\n", signo, hostname);
 	} else {
-	    mprintf("Database updated (%d signatures) from %s (%s).\n", signo, hostname, ipaddr);
-	    logg("Database updated (%d signatures) from %s (%s).\n", signo, hostname, ipaddr);
+	    mprintf("Database updated (%d signatures) from %s (IP: %s)\n", signo, hostname, ipaddr);
+	    logg("Database updated (%d signatures) from %s (IP: %s)\n", signo, hostname, ipaddr);
 	}
 
 #ifdef BUILD_CLAMD
 	if(optl(opt, "daemon-notify")) {
 		const char *clamav_conf = getargl(opt, "daemon-notify");
 	    if(!clamav_conf)
-		clamav_conf = CONFDIR"/clamav.conf";
+		clamav_conf = CONFDIR"/clamd.conf";
 
 	    notify(clamav_conf);
 	} else if((cpt = cfgopt(copt, "NotifyClamd"))) {
 		const char *clamav_conf = cpt->strarg;
 	    if(!clamav_conf)
-		clamav_conf = CONFDIR"/clamav.conf";
+		clamav_conf = CONFDIR"/clamd.conf";
 
 	    notify(clamav_conf);
 	}
@@ -111,112 +189,169 @@ int downloadmanager(const struct cfgstruct *copt, const struct optstruct *opt, c
 	return 1;
 }
 
-int downloaddb(const char *localname, const char *remotename, const char *hostname, char *ip, int *signo, const struct cfgstruct *copt)
+static int isnumb(const char *str)
+{
+	int i;
+
+    for(i = 0; i < strlen(str); i++)
+	if(!isdigit(str[i]))
+	    return 0;
+
+    return 1;
+}
+
+int downloaddb(const char *localname, const char *remotename, const char *hostname, char *ip, int *signo, const struct cfgstruct *copt, const char *dnsreply)
 {
 	struct cl_cvd *current, *remote;
 	struct cfgstruct *cpt;
-	int hostfd, nodb = 0, dbver = 0, ret, port = 0;
-	char  *tempname, ipaddr[16];
+	int hostfd, nodb = 0, dbver = -1, ret, port = 0, ims = -1;
+	char  *tempname, ipaddr[16], *pt;
 	const char *proxy = NULL, *user = NULL, *pass = NULL;
 	int flevel = cl_retflevel();
 
+
     if((current = cl_cvdhead(localname)) == NULL)
 	nodb = 1;
 
-    if((cpt = cfgopt(copt, "HTTPProxyUsername"))) {
-	user = cpt->strarg;
+    if(!nodb && dnsreply) {
+	    int field = 0;
 
-	if((cpt = cfgopt(copt, "HTTPProxyPassword"))) {
-	    pass = cpt->strarg;
+	if(!strcmp(remotename, "main.cvd")) {
+	    field = 1;
+	} else if(!strcmp(remotename, "daily.cvd")) {
+	    field = 2;
 	} else {
-	    mprintf("HTTPProxyUsername requires HTTPProxyPassword\n");
-	    return 56;
+	    mprintf("WARNING: Unknown database name (%s) passed.\n", remotename);
+	    logg("WARNING: Unknown database name (%s) passed.\n", remotename);
 	}
+
+	if(field && (pt = cli_strtok(dnsreply, field, ":"))) {
+	    if(!isnumb(pt)) {
+		mprintf("WARNING: Broken database version in TXT record.\n");
+		logg("WARNING: Broken database version in TXT record.\n");
+	    } else {
+		dbver = atoi(pt);
+		mprintf("*%s version from DNS: %d\n", remotename, dbver);
+	    }
+	    free(pt);
+	} else {
+	    mprintf("WARNING: Invalid DNS reply. Falling back to HTTP mode.\n");
+	    logg("WARNING: Invalid DNS reply. Falling back to HTTP mode.\n");
+	}
+
     }
 
-    /*
-     * njh@bandsman.co.uk: added proxy support. Tested using squid 2.4
-     */
+    /* Initialize proxy settings */
     if((cpt = cfgopt(copt, "HTTPProxyServer"))) {
 	proxy = cpt->strarg;
 	if(strncasecmp(proxy, "http://", 7) == 0)
 	    proxy += 7;
+
+	if((cpt = cfgopt(copt, "HTTPProxyUsername"))) {
+	    user = cpt->strarg;
+	    if((cpt = cfgopt(copt, "HTTPProxyPassword"))) {
+		pass = cpt->strarg;
+	    } else {
+		mprintf("HTTPProxyUsername requires HTTPProxyPassword\n");
+		if(current)
+		    cl_cvdfree(current);
+		return 56;
+	    }
+	}
+
+	if((cpt = cfgopt(copt, "HTTPProxyPort")))
+	    port = cpt->numarg;
+
 	mprintf("Connecting via %s\n", proxy);
     }
 
-    if((cpt = cfgopt(copt, "HTTPProxyPort")))
-	port = cpt->numarg;
+    memset(ipaddr, 0, sizeof(ipaddr));
 
-    if(ip[0])
-	hostfd = wwwconnect(ip, proxy, port, ipaddr); /* we use ip to connect */
-    else
-	hostfd = wwwconnect(hostname, proxy, port, ipaddr);
+    if(!nodb && dbver == -1) {
+	if(ip[0]) /* use ip to connect */
+	    hostfd = wwwconnect(ip, proxy, port, ipaddr);
+	else
+	    hostfd = wwwconnect(hostname, proxy, port, ipaddr);
 
-    if(hostfd < 0) {
-	mprintf("@Connection with %s (IP: %s) failed.\n", hostname, ipaddr);
-	return 52;
-    } else
-	mprintf("*Connected to %s (%s).\n", hostname, ipaddr);
+	if(hostfd < 0) {
+            mprintf("@No servers could be reached. Giving up\n");
+	    if(current)
+		cl_cvdfree(current);
+	    return 52;
+	} else {
+	    mprintf("*Connected to %s (IP: %s).\n", hostname, ipaddr);
+	}
 
-    if(!ip[0])
-	strcpy(ip, ipaddr);
+	if(!ip[0])
+	    strcpy(ip, ipaddr);
 
-    if(!(remote = remote_cvdhead(remotename, hostfd, hostname, proxy, user, pass))) {
-	mprintf("@Can't read %s header from %s (%s)\n", remotename, hostname, ipaddr);
+	remote = remote_cvdhead(remotename, hostfd, hostname, proxy, user, pass, &ims);
+
+	if(!nodb && !ims) {
+	    mprintf("%s is up to date (version: %d, sigs: %d, f-level: %d, builder: %s)\n", localname, current->version, current->sigs, current->fl, current->builder);
+	    logg("%s is up to date (version: %d, sigs: %d, f-level: %d, builder: %s)\n", localname, current->version, current->sigs, current->fl, current->builder);
+	    *signo += current->sigs;
+	    cl_cvdfree(current);
+	    return 1;
+	}
+
+	if(!remote) {
+	    mprintf("@Can't read %s header from %s (IP: %s)\n", remotename, hostname, ipaddr);
+	    close(hostfd);
+	    if(current)
+		cl_cvdfree(current);
+	    return 58;
+	}
+
+	dbver = remote->version;
+	cl_cvdfree(remote);
 	close(hostfd);
-	return 58;
     }
 
-    *signo += remote->sigs; /* we need to do it just here */
-
-    if(current && (current->version >= remote->version)) {
+    if(!nodb && (current->version >= dbver)) {
 	mprintf("%s is up to date (version: %d, sigs: %d, f-level: %d, builder: %s)\n", localname, current->version, current->sigs, current->fl, current->builder);
 	logg("%s is up to date (version: %d, sigs: %d, f-level: %d, builder: %s)\n", localname, current->version, current->sigs, current->fl, current->builder);
 
 	if(flevel < current->fl) {
 	    /* display warning even for already installed database */
-	    mprintf("WARNING: Your ClamAV installation is OUTDATED - please update immediately !\n");
+	    mprintf("WARNING: Your ClamAV installation is OUTDATED - please update immediately!\n");
 	    mprintf("WARNING: Current functionality level = %d, required = %d\n", flevel, current->fl);
-	    logg("WARNING: Your ClamAV installation is OUTDATED - please update immediately !\n");
+	    logg("WARNING: Your ClamAV installation is OUTDATED - please update immediately!\n");
 	    logg("WARNING: Current functionality level = %d, required = %d\n", flevel, current->fl);
 	}
 
-	close(hostfd);
+	*signo += current->sigs;
 	cl_cvdfree(current);
-	cl_cvdfree(remote);
 	return 1;
     }
 
-    dbver = remote->version;
-    
     if(current)
 	cl_cvdfree(current);
 
-    cl_cvdfree(remote);
-
-    /* FIXME: We need to reconnect, because we may not be able to download
-     * the database. The problem doesn't exist with my local apache.
-     * Some code change is needed in get_md5_checksum().
-     */
-    /* begin bug work-around */
-    close(hostfd);
-    hostfd = wwwconnect(ipaddr, proxy, port, NULL); /* we use ipaddr to connect
-					       * to the same mirror
-					       */
+    if(ipaddr[0]) {
+	/* use ipaddr in order to connect to the same mirror */
+	hostfd = wwwconnect(ipaddr, proxy, port, NULL);
+    } else {
+	hostfd = wwwconnect(hostname, proxy, port, ipaddr);
+	if(!ip[0])
+	    strcpy(ip, ipaddr);
+    }
 
     if(hostfd < 0) {
-	mprintf("@Connection with %s failed.\n", ipaddr);
+	if(ipaddr[0])
+	    mprintf("Connection with %s (IP: %s) failed.\n", hostname, ipaddr);
+	else
+	    mprintf("Connection with %s failed.\n", hostname);
 	return 52;
     };
-    /* end */
 
-    /* temporary file is created in clamav's directory thus we don't need
-     * to create it immediately because race condition is not possible here
+    /* the temporary file is created in a directory owned by clamav so race
+     * conditions are not possible
      */
-    tempname = cl_gentemp(".");
+    tempname = cli_gentemp(".");
 
     if((ret = get_database(remotename, hostfd, tempname, hostname, proxy, user, pass))) {
-        mprintf("@Can't download %s from %s\n", remotename, ipaddr);
+        mprintf("@Can't download %s from %s (IP: %s)\n", remotename, hostname, ipaddr);
         unlink(tempname);
         free(tempname);
         close(hostfd);
@@ -233,7 +368,7 @@ int downloaddb(const char *localname, const char *remotename, const char *hostna
     }
 
     if((current = cl_cvdhead(tempname)) == NULL) {
-	mprintf("@Can't read CVD header of new %s database.\n", localname); /* we lie :) */
+	mprintf("@Can't read CVD header of new %s database.\n", localname);
 	unlink(tempname);
 	free(tempname);
 	return 54;
@@ -261,12 +396,13 @@ int downloaddb(const char *localname, const char *remotename, const char *hostna
     logg("%s updated (version: %d, sigs: %d, f-level: %d, builder: %s)\n", localname, current->version, current->sigs, current->fl, current->builder);
 
     if(flevel < current->fl) {
-	mprintf("WARNING: Your ClamAV installation is OUTDATED - please update immediately !\n");
+	mprintf("WARNING: Your ClamAV installation is OUTDATED - please update immediately!\n");
 	mprintf("WARNING: Current functionality level = %d, required = %d\n", flevel, current->fl);
-	logg("WARNING: Your ClamAV installation is OUTDATED - please update immediately !\n");
+	logg("WARNING: Your ClamAV installation is OUTDATED - please update immediately!\n");
 	logg("WARNING: Current functionality level = %d, required = %d\n", flevel, current->fl);
     }
 
+    *signo += current->sigs;
     cl_cvdfree(current);
     free(tempname);
     return 0;
@@ -276,7 +412,7 @@ int downloaddb(const char *localname, const char *remotename, const char *hostna
 /* proxy support finshed by njh@bandsman.co.uk */
 int wwwconnect(const char *server, const char *proxy, int pport, char *ip)
 {
-	int socketfd, port;
+	int socketfd = -1, port, i;
 	struct sockaddr_in name;
 	struct hostent *host;
 	char ipaddr[16];
@@ -287,13 +423,6 @@ int wwwconnect(const char *server, const char *proxy, int pport, char *ip)
     if(ip)
 	strcpy(ip, "???");
 
-    /* njh@bandsman.co.uk: for BEOS */
-#ifdef PF_INET
-    socketfd = socket(PF_INET, SOCK_STREAM, 0);
-#else
-    socketfd = socket(AF_INET, SOCK_STREAM, 0);
-#endif
-
     name.sin_family = AF_INET;
 
     if(proxy) {
@@ -324,33 +453,55 @@ int wwwconnect(const char *server, const char *proxy, int pport, char *ip)
 	return -1;
     }
 
-    /* this dirty hack comes from pink - Nosuid TCP/IP ping 1.6 */
-    ia = (unsigned char *) host->h_addr;
-    sprintf(ipaddr, "%u.%u.%u.%u", ia[0], ia[1], ia[2], ia[3]);
+    for(i = 0; host->h_addr_list[i] != 0; i++) {
+	/* this dirty hack comes from pink - Nosuid TCP/IP ping 1.6 */
+	ia = (unsigned char *) host->h_addr_list[i];
+	sprintf(ipaddr, "%u.%u.%u.%u", ia[0], ia[1], ia[2], ia[3]);
 
-    if(ip)
-	strcpy(ip, ipaddr);
+	if(ip)
+	    strcpy(ip, ipaddr);
 
-    name.sin_addr = *((struct in_addr *) host->h_addr);
-    name.sin_port = htons(port);
+	if(i > 0)
+	    mprintf("Trying host %s (%s)...\n", hostpt, ipaddr);
 
-    if(connect(socketfd, (struct sockaddr *) &name, sizeof(struct sockaddr_in)) == -1) {
-	mprintf("@Can't connect to port %d of host %s (%s)\n", port, hostpt, ipaddr);
-	close(socketfd);
-	return -2;
+	name.sin_addr = *((struct in_addr *) host->h_addr_list[i]);
+	name.sin_port = htons(port);
+
+#ifdef PF_INET
+	socketfd = socket(PF_INET, SOCK_STREAM, 0);
+#else
+	socketfd = socket(AF_INET, SOCK_STREAM, 0);
+#endif
+
+	if(connect(socketfd, (struct sockaddr *) &name, sizeof(struct sockaddr_in)) == -1) {
+	    mprintf("Can't connect to port %d of host %s (IP: %s)\n", port, hostpt, ipaddr);
+	    close(socketfd);
+	    continue;
+	}
+
+	return socketfd;
     }
 
-    return socketfd;
+    return -2;
 }
 
-/* njh@bandsman.co.uk: added proxy support */
-/* TODO: use a HEAD instruction to see if the file has been changed */
-struct cl_cvd *remote_cvdhead(const char *file, int socketfd, const char *hostname, const char *proxy, const char *user, const char *pass)
+int Rfc2822DateTime(char *buf, time_t mtime)
+{
+	struct tm *time;
+
+    time = gmtime(&mtime);
+    return strftime(buf, 36, "%a, %d %b %Y %X GMT", time);
+}
+
+struct cl_cvd *remote_cvdhead(const char *file, int socketfd, const char *hostname, const char *proxy, const char *user, const char *pass, int *ims)
 {
 	char cmd[512], head[513], buffer[FILEBUFF], *ch, *tmp;
 	int i, j, bread, cnt;
 	char *remotename = NULL, *authorization = NULL;
 	struct cl_cvd *cvd;
+	char last_modified[36];
+	struct stat sb;
+
 
     if(proxy) {
         remotename = mmalloc(strlen(hostname) + 8);
@@ -370,24 +521,33 @@ struct cl_cvd *remote_cvdhead(const char *file, int socketfd, const char *hostna
         }
     }
 
-    mprintf("Reading CVD header (%s): ", file);
+    if(stat(file, &sb) != -1 && sb.st_mtime < time(NULL)) {
+	Rfc2822DateTime(last_modified, sb.st_mtime);
+    } else {
+	    time_t mtime = 1104119530;
+	Rfc2822DateTime(last_modified, mtime);
+	mprintf("*Assuming modification time in the past\n");
+    }
 
+    mprintf("*If-Modified-Since: %s\n", last_modified);
+
+    mprintf("Reading CVD header (%s): ", file);
 #ifdef	NO_SNPRINTF
     sprintf(cmd, "GET %s/%s HTTP/1.1\r\n"
 	"Host: %s\r\n%s"
 	"User-Agent: "PACKAGE"/"VERSION"\r\n"
-	"Cache-Control: no-cache\r\n"
 	"Connection: close\r\n"
 	"Range: bytes=0-511\r\n"
-	"\r\n", (remotename != NULL)?remotename:"", file, hostname, (authorization != NULL)?authorization:"");
+        "If-Modified-Since: %s\r\n"
+        "\r\n", (remotename != NULL)?remotename:"", file, hostname, (authorization != NULL)?authorization:"", last_modified);
 #else
     snprintf(cmd, sizeof(cmd), "GET %s/%s HTTP/1.1\r\n"
 	"Host: %s\r\n%s"
 	"User-Agent: "PACKAGE"/"VERSION"\r\n"
-	"Cache-Control: no-cache\r\n"
 	"Connection: close\r\n"
 	"Range: bytes=0-511\r\n"
-	"\r\n", (remotename != NULL)?remotename:"", file, hostname, (authorization != NULL)?authorization:"");
+        "If-Modified-Since: %s\r\n"
+        "\r\n", (remotename != NULL)?remotename:"", file, hostname, (authorization != NULL)?authorization:"", last_modified);
 #endif
     write(socketfd, cmd, strlen(cmd));
 
@@ -407,9 +567,19 @@ struct cl_cvd *remote_cvdhead(const char *file, int socketfd, const char *hostna
 	return NULL;
     }
 
-    if ((strstr(buffer, "HTTP/1.1 404")) != NULL) { 
-      mprintf("@CVD file not found on remote server\n");
-      return NULL;
+    if((strstr(buffer, "HTTP/1.1 404")) != NULL || (strstr(buffer, "HTTP/1.0 404")) != NULL) { 
+	mprintf("@CVD file not found on remote server\n");
+	return NULL;
+    }
+
+    /* check whether the resource is up-to-date */
+    if((strstr(buffer, "HTTP/1.1 304")) != NULL || (strstr(buffer, "HTTP/1.0 304")) != NULL) { 
+
+	*ims = 0;
+	mprintf("OK (IMS)\n");
+	return NULL;
+    } else {
+	*ims = 1;
     }
 
     ch = buffer;
@@ -442,8 +612,6 @@ struct cl_cvd *remote_cvdhead(const char *file, int socketfd, const char *hostna
     return cvd;
 }
 
-/* njh@bandsman.co.uk: added proxy support */
-/* TODO: use a HEAD instruction to see if the file has been changed */
 int get_database(const char *dbfile, int socketfd, const char *file, const char *hostname, const char *proxy, const char *user, const char *pass)
 {
 	char cmd[512], buffer[FILEBUFF], *ch;
@@ -470,13 +638,16 @@ int get_database(const char *dbfile, int socketfd, const char *file, const char
         }
     }
 
-#ifdef C_CYGWIN
+#if defined(C_CYGWIN) || defined(C_OS2)
     if((fd = open(file, O_WRONLY|O_CREAT|O_EXCL|O_BINARY, 0644)) == -1) {
 #else
     if((fd = open(file, O_WRONLY|O_CREAT|O_EXCL, 0644)) == -1) {
 #endif
-	mprintf("@Can't open new file %s to write\n", file);
-	perror("open");
+	    char currdir[512];
+
+	getcwd(currdir, sizeof(currdir));
+	mprintf("@Can't create new file %s in %s\n", file, currdir);
+	mprintf("@The database directory must be writable for UID %d or GID %d\n", getuid(), getgid());
 	return 57;
     }
 
@@ -500,21 +671,21 @@ int get_database(const char *dbfile, int socketfd, const char *file, const char
     free(remotename);
     free(authorization);
 
-    if ((bread = recv(socketfd, buffer, FILEBUFF, 0)) == -1) {
-      mprintf("@Error while reading database from %s\n", hostname);
-      return 52;
-    }
-
-    if ((strstr(buffer, "HTTP/1.1 404")) != NULL) { 
-      mprintf("@%s not found on remote server\n", dbfile);
-      return 58;
-    }
+    /* read all the http headers */
 
     ch = buffer;
     i = 0;
     while (1) {
-      if (*ch == '\n' && *(ch - 1) == '\r' && *(ch - 2) == '\n' && *(ch - 3) == '\r') {
-	ch++;
+      /* recv one byte at a time, until we reach \r\n\r\n */
+
+      if(recv(socketfd, buffer + i, 1, 0) == -1) {
+        mprintf("@Error while reading database from %s\n", hostname);
+        close(fd);
+        unlink(file);
+        return 52;
+      }
+
+      if (i>2 && *ch == '\n' && *(ch - 1) == '\r' && *(ch - 2) == '\n' && *(ch - 3) == '\r') {
 	i++;
 	break;
       }
@@ -522,7 +693,19 @@ int get_database(const char *dbfile, int socketfd, const char *file, const char
       i++;
     }
 
-    write(fd, ch, bread - i);
+    buffer[i] = 0;
+
+    /* check whether the resource actually existed or not */
+
+    if ((strstr(buffer, "HTTP/1.1 404")) != NULL) { 
+      mprintf("@%s not found on remote server\n", dbfile);
+      close(fd);
+      unlink(file);
+      return 58;
+    }
+
+    /* receive body and write it to disk */
+
     while((bread = read(socketfd, buffer, FILEBUFF))) {
 	write(fd, buffer, bread);
 	mprintf("Downloading %s [%c]\r", dbfile, rotation[rot]);
