@@ -29,6 +29,7 @@
 #include <sys/stat.h>
 #include <fcntl.h>
 #include <zlib.h>
+#include <time.h>
 
 #include "clamav.h"
 #include "others.h"
@@ -124,6 +125,7 @@ int cli_untgz(int fd, const char *destdir)
 		cli_errmsg("Invalid size in header.\n");
 		free(fullname);
 	        gzclose(infile);
+		fclose(outfile);
 		return -1;
 	    }
 
@@ -221,6 +223,13 @@ struct cl_cvd *cl_cvdparse(const char *head)
 	return NULL;
     }
 
+    if((pt = cli_strtok(head, 8, ":"))) {
+	cvd->stime = atoi(pt);
+	free(pt);
+    } else
+	cli_dbgmsg("CVD -> No creation time in seconds (old file format)\n");
+
+
     return cvd;
 }
 
@@ -235,7 +244,7 @@ struct cl_cvd *cl_cvdhead(const char *file)
 	return NULL;
     }
 
-    if((i=fread(head, 1, 512, fd)) != 512) {
+    if((i = fread(head, 1, 512, fd)) != 512) {
 	cli_dbgmsg("Short read (%d) while reading CVD head from %s\n", i, file);
 	fclose(fd);
 	return NULL;
@@ -258,7 +267,7 @@ void cl_cvdfree(struct cl_cvd *cvd)
     free(cvd);
 }
 
-int cli_cvdverify(FILE *fd)
+int cli_cvdverify(FILE *fd, struct cl_cvd *cvdpt)
 {
 	struct cl_cvd *cvd;
 	char *md5, head[513];
@@ -276,8 +285,10 @@ int cli_cvdverify(FILE *fd)
     if((cvd = cl_cvdparse(head)) == NULL)
 	return CL_ECVD;
 
+    if(cvdpt)
+	memcpy(cvdpt, cvd, sizeof(struct cl_cvd));
 
-    md5 = cli_md5stream(fd);
+    md5 = cli_md5stream(fd, NULL);
     cli_dbgmsg("MD5(.tar.gz) = %s\n", md5);
 
     if(strncmp(md5, cvd->md5, 32)) {
@@ -311,38 +322,48 @@ int cl_cvdverify(const char *file)
 	return CL_EOPEN;
     }
 
-    ret = cli_cvdverify(fd);
+    ret = cli_cvdverify(fd, NULL);
     fclose(fd);
 
     return ret;
 }
 
-int cli_cvdload(FILE *fd, struct cl_node **root, int *virnum)
+int cli_cvdload(FILE *fd, struct cl_node **root, unsigned int *signo, short warn)
 {
         char *dir, *tmp, *buffer;
+	struct cl_cvd cvd;
 	int bytes, ret;
-	const char *tmpdir;
 	FILE *tmpd;
+	time_t stime;
+
 
     cli_dbgmsg("in cli_cvdload()\n");
 
     /* verify */
 
-    if((ret = cli_cvdverify(fd)))
+    if((ret = cli_cvdverify(fd, &cvd)))
 	return ret;
 
+    if(cvd.stime && warn) {
+	time(&stime);
+	if((int) stime - cvd.stime > 604800) {
+	    cli_warnmsg("**************************************************\n");
+	    cli_warnmsg("***  The virus database is older than 7 days.  ***\n");
+	    cli_warnmsg("***        Please update it IMMEDIATELY!       ***\n");
+	    cli_warnmsg("**************************************************\n");
+	}
+    }
+
+    if(cvd.fl > cl_retflevel()) {
+	cli_warnmsg("****************************************************\n");
+	cli_warnmsg("***  This version of ClamAV engine is outdated.  ***\n");
+	cli_warnmsg("***         Please update it IMMEDIATELY!        ***\n");
+	cli_warnmsg("****************************************************\n");
+    }
+
     fseek(fd, 512, SEEK_SET);
 
-    tmpdir = getenv("TMPDIR");
-
-    if(tmpdir == NULL)
-#ifdef P_tmpdir
-	tmpdir = P_tmpdir;
-#else
-	tmpdir = "/tmp";
-#endif
-
-    dir = cl_gentemp(tmpdir);
+    dir = cli_gentemp(NULL);
     if(mkdir(dir, 0700)) {
 	cli_errmsg("cli_cvdload():  Can't create temporary directory %s\n", dir);
 	return CL_ETMPDIR;
@@ -361,7 +382,7 @@ int cli_cvdload(FILE *fd, struct cl_node **root, int *virnum)
 
 	    /* start */
 
-	    tmp = cl_gentemp(tmpdir);
+	    tmp = cli_gentemp(NULL);
 	    if((tmpd = fopen(tmp, "wb+")) == NULL) {
 		cli_errmsg("Can't create temporary file %s\n", tmp);
 		free(dir);
@@ -369,8 +390,12 @@ int cli_cvdload(FILE *fd, struct cl_node **root, int *virnum)
 		return CL_ETMPFILE;
 	    }
 
-	    if(!(buffer = (char *) cli_malloc(FILEBUFF)))
+	    if(!(buffer = (char *) cli_malloc(FILEBUFF))) {
+		free(dir);
+		free(tmp);
+		fclose(tmpd);
 		return CL_EMEM;
+	    }
 
 	    while((bytes = fread(buffer, 1, FILEBUFF, fd)) > 0)
 		fwrite(buffer, 1, bytes, tmpd);
@@ -384,6 +409,7 @@ int cli_cvdload(FILE *fd, struct cl_node **root, int *virnum)
 		cli_errmsg("cli_cvdload(): Can't unpack CVD file.\n");
 		cli_rmdirs(dir);
 		free(dir);
+		fclose(tmpd);
 		unlink(tmp);
 		free(tmp);
 		return CL_ECVDEXTR;
@@ -396,7 +422,7 @@ int cli_cvdload(FILE *fd, struct cl_node **root, int *virnum)
 	    /* end */
 
     /* load extracted directory */
-    cl_loaddbdir(dir, root, virnum);
+    cl_loaddbdir(dir, root, signo);
 
     cli_rmdirs(dir);
     free(dir);
