@@ -461,7 +461,7 @@ static void ole2_walk_property_tree(int fd, ole2_header_t *hdr, const char *dir,
 				int rec_level, int *file_count, const struct cl_limits *limits)
 {
 	property_t prop_block[4];
-	int32_t index, current_block, count=0, i;
+	int32_t index, current_block, i;
 	unsigned char *dirname;
 	current_block = hdr->prop_start;
 	
@@ -469,11 +469,16 @@ static void ole2_walk_property_tree(int fd, ole2_header_t *hdr, const char *dir,
 		return;
 	}
 	
-	if(limits && limits->maxfiles && (*file_count > limits->maxfiles)) {
+	if (limits && limits->maxfiles && (*file_count > limits->maxfiles)) {
 		cli_dbgmsg("OLE2: File limit reached (max: %d)\n", limits->maxfiles);
 		return;
 	}
 	
+	if (limits && limits->maxreclevel && (rec_level > limits->maxreclevel)) {
+		cli_dbgmsg("OLE2: Recursion limit reached (max: %d)\n", limits->maxreclevel);
+		return;
+	}
+	
 	index = prop_index / 4;
 	for (i=0 ; i < index ; i++) {
 		current_block = ole2_get_next_block_number(fd, hdr, current_block);
@@ -563,7 +568,7 @@ static void ole2_walk_property_tree(int fd, ole2_header_t *hdr, const char *dir,
 /* Write file Handler - write the contents of the entry to a file */
 static int handler_writefile(int fd, ole2_header_t *hdr, property_t *prop, const char *dir)
 {
-	unsigned char buff[(1 << hdr->log2_big_block_size)];
+	unsigned char *buff;
 	int32_t current_block, ofd, len, offset;
 	char *name, *newname;
 
@@ -579,16 +584,30 @@ static int handler_writefile(int fd, ole2_header_t *hdr, property_t *prop, const
 
 	if (! (name = get_property_name(prop->name, prop->name_size))) {
 		/* File without a name - create a name for it */
-		int i;
+		off_t i;
                                                                                                                             
 		i = lseek(fd, 0, SEEK_CUR);
 		name = (char *) cli_malloc(11);
 		if (!name) {
 			return FALSE;
 		}
-		snprintf(name, 11, "%.10d", i + (int) prop);
+		snprintf(name, 11, "%.10ld", i + (long int) prop);
+	} else {
+		/* Sanitize the file name */
+                for(newname = name; *newname; newname++) {
+#ifdef  C_DARWIN
+                        *newname &= '\177';
+#endif
+#if     defined(MSDOS) || defined(C_CYGWIN) || defined(WIN32) || defined(C_OS2)
+                        if(strchr("/*?<>|\"+=,;: ", *newname))
+#else
+                        if(*newname == '/')
+#endif
+                                *newname = '_';
+                }
 	}
 
+
 	newname = (char *) cli_malloc(strlen(name) + strlen(dir) + 2);
 	if (!newname) {
 		free(name);
@@ -599,24 +618,34 @@ static int handler_writefile(int fd, ole2_header_t *hdr, property_t *prop, const
 
 	ofd = open(newname, O_WRONLY|O_CREAT|O_TRUNC, S_IRWXU);
 	if (ofd < 0) {
+		cli_errmsg("ERROR: failed to create file: %s\n", newname);
+		free(newname);
 		return FALSE;
 	}
 	free(newname);
 	current_block = prop->start_block;
 	len = prop->size;
 
+	buff = (unsigned char *) cli_malloc(1 << hdr->log2_big_block_size);
+	if (!buff) {
+		close(ofd);
+		return FALSE;
+	}
+
 	while((current_block >= 0) && (len > 0)) {
 		if (prop->size < hdr->sbat_cutoff) {
 			/* Small block file */
-			if (!ole2_get_sbat_data_block(fd, hdr, &buff, current_block)) {
+			if (!ole2_get_sbat_data_block(fd, hdr, buff, current_block)) {
 				cli_dbgmsg("ole2_get_sbat_data_block failed\n");
 				close(ofd);
+				free(buff);
 				return FALSE;
 			}
 			/* buff now contains the block with 8 small blocks in it */
 			offset = 64 * (current_block % 8);
 			if (cli_writen(ofd, &buff[offset], MIN(len,64)) != MIN(len,64)) {
 				close(ofd);
+				free(buff);
 				return FALSE;
 			}
 
@@ -624,13 +653,15 @@ static int handler_writefile(int fd, ole2_header_t *hdr, property_t *prop, const
 			current_block = ole2_get_next_sbat_block(fd, hdr, current_block);
 		} else {
 			/* Big block file */
-			if (!ole2_read_block(fd, hdr, &buff, current_block)) {
+			if (!ole2_read_block(fd, hdr, buff, current_block)) {
 				close(ofd);
+				free(buff);
 				return FALSE;
 			}
-			if (cli_writen(ofd, &buff, MIN(len,(1 << hdr->log2_big_block_size))) !=
+			if (cli_writen(ofd, buff, MIN(len,(1 << hdr->log2_big_block_size))) !=
 							MIN(len,(1 << hdr->log2_big_block_size))) {
 				close(ofd);
+				free(buff);
 				return FALSE;
 			}
 
@@ -639,6 +670,7 @@ static int handler_writefile(int fd, ole2_header_t *hdr, property_t *prop, const
 		}
 	}
 	close(ofd);
+	free(buff);
 	return TRUE;
 }
 
@@ -762,6 +794,11 @@ int cli_ole2_extract(int fd, const char *dirname, const struct cl_limits *limits
 	
 	if (strncmp(hdr.magic, magic_id, 8) != 0) {
 		cli_dbgmsg("OLE2 magic failed!\n");
+#ifdef HAVE_MMAP
+		if (hdr.m_area != NULL) {
+			munmap(hdr.m_area, hdr.m_length);
+		}
+#endif
 		return CL_EOLE2;
 	}
 
