@@ -1,5 +1,5 @@
 /*
- *  Copyright (C) 1999 - 2004 Tomasz Kojm <tk@clamav.net>
+ *  Copyright (C) 1999 - 2005 Tomasz Kojm <tkojm@clamav.net>
  *
  *  This program is free software; you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License as published by
@@ -38,18 +38,41 @@
 #include <errno.h>
 #include <target.h>
 #include <sys/time.h>
+#include <sys/param.h>
 
 #ifdef CL_THREAD_SAFE
 #  include <pthread.h>
-pthread_mutex_t cl_gentemp_mutex = PTHREAD_MUTEX_INITIALIZER;
+pthread_mutex_t cli_gentemp_mutex = PTHREAD_MUTEX_INITIALIZER;
+#endif
+
+#if defined(HAVE_READDIR_R_3) || defined(HAVE_READDIR_R_2)
+#include <limits.h>
+#include <stddef.h>
 #endif
 
 #include "clamav.h"
 #include "others.h"
 #include "md5.h"
+#include "cltypes.h"
 
-#define CL_FLEVEL 2 /* don't touch it */
+/* Maximum filenames under various systems - njh */
+#ifndef	NAME_MAX	/* e.g. Linux */
+# ifdef	MAXNAMELEN	/* e.g. Solaris */
+#   define	NAME_MAX	MAXNAMELEN
+# else
+#   ifdef	FILENAME_MAX	/* e.g. SCO */
+#     define	NAME_MAX	FILENAME_MAX
+#   endif
+# endif
+#endif
 
+#ifndef	O_BINARY
+#define	O_BINARY	0
+#endif
+
+#define CL_FLEVEL 4 /* don't touch it */
+
+#define MAX_ALLOCATION 134217728
 
 short cli_debug_flag = 0, cli_leavetemps_flag = 0;
 
@@ -99,6 +122,11 @@ int cl_retflevel(void)
     return CL_FLEVEL;
 }
 
+const char *cl_retver(void)
+{
+    return VERSION;
+}
+
 const char *cl_strerror(int clerror)
 {
     switch(clerror) {
@@ -152,6 +180,8 @@ const char *cl_strerror(int clerror)
 	    return "Null argument passed while initialized is required";
 	case CL_EIO:
 	    return "Input/Output error";
+	case CL_EFORMAT:
+	    return "Bad format or broken data";
 	default:
 	    return "Unknown error code";
     }
@@ -162,70 +192,89 @@ const char *cl_perror(int clerror)
     return cl_strerror(clerror);
 }
 
-char *cl_md5file(const char *filename)
+char *cli_md5stream(FILE *fs, unsigned char *digcpy)
 {
-	FILE *fd;
-	unsigned char buffer[16];
-	char *md5str;
-	int i, cnt=0;
+	unsigned char digest[16];
+	char buff[FILEBUFF];
+	struct MD5Context ctx;
+	char *md5str, *pt;
+	int i, bytes;
 
-    if((fd = fopen(filename, "rb")) == NULL) {
-	cli_errmsg("md5_file(): Can't read file %s\n", filename);
+
+    MD5Init(&ctx);
+
+    while((bytes = fread(buff, 1, FILEBUFF, fs)))
+	MD5Update(&ctx, buff, bytes);
+
+    MD5Final(digest, &ctx);
+
+    if(!(md5str = (char *) cli_calloc(32 + 1, sizeof(char))))
+	return NULL;
+
+    pt = md5str;
+    for(i = 0; i < 16; i++) {
+	sprintf(pt, "%02x", digest[i]);
+	pt += 2;
+    }
+
+    if(digcpy)
+	memcpy(digcpy, digest, 16);
+
+    return md5str;
+}
+
+char *cli_md5file(const char *filename)
+{
+	FILE *fs;
+	char *md5str;
+
+
+    if((fs = fopen(filename, "rb")) == NULL) {
+	cli_errmsg("cli_md5file(): Can't read file %s\n", filename);
 	return NULL;
     }
 
-    md5_stream(fd, &buffer);
-    fclose(fd);
+    md5str = cli_md5stream(fs, NULL);
+    fclose(fs);
 
-    md5str = (char*) calloc(32 + 1, sizeof(char));
-
-    for(i=0; i<16; i++)
-	cnt += sprintf(md5str + cnt, "%02x", buffer[i]);
-
-    return(md5str);
+    return md5str;
 }
 
-char *cli_md5stream(FILE *fd)
+static char *cli_md5buff(const char *buffer, unsigned int len)
 {
-	unsigned char buffer[16];
-	char *md5str;
-	int i, cnt=0;
-
-    md5_stream(fd, &buffer);
-
-    md5str = (char*) calloc(32 + 1, sizeof(char));
-
-    for(i=0; i<16; i++)
-	cnt += sprintf(md5str + cnt, "%02x", buffer[i]);
-
-    return(md5str);
-}
-
-char *cl_md5buff(const char *buffer, unsigned int len)
-{
-	unsigned char md5buff[16];
-	char *md5str;
-	struct md5_ctx ctx;
-	int i, cnt=0;
+	unsigned char digest[16];
+	char *md5str, *pt;
+	struct MD5Context ctx;
+	int i;
 
 
-    md5_init_ctx(&ctx);
-    md5_process_bytes(buffer, len, &ctx);
-    md5_finish_ctx(&ctx, &md5buff);
-    memcpy(oldmd5buff, md5buff, 16);
+    MD5Init(&ctx);
+    MD5Update(&ctx, buffer, len);
+    MD5Final(digest, &ctx);
+    memcpy(oldmd5buff, digest, 16);
 
-    md5str = (char*) cli_calloc(32 + 1, sizeof(char));
+    if(!(md5str = (char *) cli_calloc(32 + 1, sizeof(char))))
+	return NULL;
 
-    for(i=0; i<16; i++)
-	cnt += sprintf(md5str + cnt, "%02x", md5buff[i]);
+    pt = md5str;
+    for(i = 0; i < 16; i++) {
+	sprintf(pt, "%02x", digest[i]);
+	pt += 2;
+    }
 
-    return(md5str);
+    return md5str;
 }
 
 void *cli_malloc(size_t size)
 {
 	void *alloc;
 
+
+    if(size > MAX_ALLOCATION) {
+	cli_errmsg("Attempt to allocate %d bytes. Please report to bugs@clamav.net\n", size);
+	return NULL;
+    }
+
     alloc = malloc(size);
 
     if(!alloc) {
@@ -240,6 +289,12 @@ void *cli_calloc(size_t nmemb, size_t size)
 {
 	void *alloc;
 
+
+    if(size > MAX_ALLOCATION) {
+	cli_errmsg("Attempt to allocate %d bytes. Please report to bugs@clamav.net\n", size);
+	return NULL;
+    }
+
     alloc = calloc(nmemb, size);
 
     if(!alloc) {
@@ -254,6 +309,7 @@ void *cli_realloc(void *ptr, size_t size)
 {
 	void *alloc;
 
+
     alloc = realloc(ptr, size);
 
     if(!alloc) {
@@ -263,7 +319,7 @@ void *cli_realloc(void *ptr, size_t size)
     } else return alloc;
 }
 
-unsigned int cl_rndnum(unsigned int max)
+unsigned int cli_rndnum(unsigned int max)
 {
     struct timeval tv;
 
@@ -291,7 +347,7 @@ void cl_settempdir(const char *dir, short leavetemps)
     cli_leavetemps_flag = leavetemps;
 }
 
-char *cl_gentemp(const char *dir)
+char *cli_gentemp(const char *dir)
 {
 	char *name, *tmp;
         const char *mdir;
@@ -300,35 +356,40 @@ char *cl_gentemp(const char *dir)
 	struct stat foo;
 
 
-    if(!dir)
-	mdir = "/tmp";
-    else
-	mdir = (char *) dir;
+    if(!dir) {
+	if((mdir = getenv("TMPDIR")) == NULL)
+#ifdef P_tmpdir
+	    mdir = P_tmpdir;
+#else
+	    mdir = "/tmp";
+#endif
+    } else
+	mdir = dir;
 
     name = (char*) cli_calloc(strlen(mdir) + 1 + 16 + 1 + 7, sizeof(char));
     if(name == NULL) {
-	cli_dbgmsg("cl_gentemp('%s'): out of memory\n", dir);
+	cli_dbgmsg("cli_gentemp('%s'): out of memory\n", dir);
 	return NULL;
     }
 
 #ifdef CL_THREAD_SAFE
-    pthread_mutex_lock(&cl_gentemp_mutex);
+    pthread_mutex_lock(&cli_gentemp_mutex);
 #endif
 
     memcpy(salt, oldmd5buff, 16);
 
     do {
 	for(i = 16; i < 48; i++)
-	    salt[i] = cl_rndnum(255);
+	    salt[i] = cli_rndnum(255);
 
-	tmp = cl_md5buff(( char* ) salt, 48);
+	tmp = cli_md5buff(( char* ) salt, 48);
 	sprintf(name, "%s/clamav-", mdir);
 	strncat(name, tmp, 16);
 	free(tmp);
     } while(stat(name, &foo) != -1);
 
 #ifdef CL_THREAD_SAFE
-    pthread_mutex_unlock(&cl_gentemp_mutex);
+    pthread_mutex_unlock(&cli_gentemp_mutex);
 #endif
 
     return(name);
@@ -338,20 +399,32 @@ int cli_rmdirs(const char *dirname)
 {
 	DIR *dd;
 	struct dirent *dent;
+#if defined(HAVE_READDIR_R_3) || defined(HAVE_READDIR_R_2)
+	union {
+	    struct dirent d;
+	    char b[offsetof(struct dirent, d_name) + NAME_MAX + 1];
+	} result;
+#endif
 	struct stat maind, statbuf;
 	char *fname;
 
 
-    if(cli_leavetemps_flag)
-	return 0;
-
     chmod(dirname, 0700);
     if((dd = opendir(dirname)) != NULL) {
 	while(stat(dirname, &maind) != -1) {
 	    if(!rmdir(dirname)) break;
 
+#ifdef HAVE_READDIR_R_3
+	    while(!readdir_r(dd, &result.d, &dent) && dent) {
+#elif defined(HAVE_READDIR_R_2)
+	    while((dent = (struct dirent *) readdir_r(dd, &result.d))) {
+#else
 	    while((dent = readdir(dd))) {
-		if(dent->d_ino) {
+#endif
+#ifndef C_INTERIX
+		if(dent->d_ino)
+#endif
+		{
 		    if(strcmp(dent->d_name, ".") && strcmp(dent->d_name, "..")) {
 			fname = cli_calloc(strlen(dirname) + strlen(dent->d_name) + 2, sizeof(char));
 			sprintf(fname, "%s/%s", dirname, dent->d_name);
@@ -450,3 +523,52 @@ int cli_writen(int fd, void *buff, unsigned int count)
         return count;
 }
 
+int32_t cli_readint32(const char *buff)
+{
+	int32_t ret;
+
+#if WORDS_BIGENDIAN == 0
+    ret = *(int32_t *) buff;
+#else
+    ret = buff[0] & 0xff;
+    ret |= (buff[1] & 0xff) << 8;
+    ret |= (buff[2] & 0xff) << 16;
+    ret |= (buff[3] & 0xff) << 24;
+#endif
+
+    return ret;
+}
+
+void cli_writeint32(char *offset, uint32_t value)
+{
+    offset[0] = value & 0xff;
+    offset[1] = (value & 0xff00) >> 8;
+    offset[2] = (value & 0xff0000) >> 16;
+    offset[3] = (value & 0xff000000) >> 24;
+}
+
+int cli_filecopy(const char *src, const char *dest)
+{
+	char *buffer;
+	int s, d, bytes;
+
+
+    if((s = open(src, O_RDONLY)) == -1)
+	return -1;
+
+    if((d = open(dest, O_CREAT|O_WRONLY|O_TRUNC|O_BINARY)) == -1) {
+	close(s);
+	return -1;
+    }
+
+    if(!(buffer = cli_malloc(FILEBUFF)))
+	return -1;
+
+    while((bytes = cli_readn(s, buffer, FILEBUFF)) > 0)
+	cli_writen(d, buffer, bytes);
+
+    free(buffer);
+    close(s);
+
+    return close(d);
+}
