@@ -1,5 +1,5 @@
 /*
- *  Copyright (C) 2002, 2003 Tomasz Kojm <zolw@konarski.edu.pl>
+ *  Copyright (C) 2002 - 2005 Tomasz Kojm <tkojm@clamav.net>
  *
  *  This program is free software; you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License as published by
@@ -46,16 +46,16 @@ static int cli_hex2int(int c)
     return -1;
 }
 
-short int *cl_hex2str(const char *hex)
+short int *cli_hex2si(const char *hex)
 {
 	short int *str, *ptr, val, c;
 	int i, len;
 
+
     len = strlen(hex);
 
-    /* additional check - hex strings are parity length here */
     if(len % 2 != 0) {
-	cli_errmsg("cl_hex2str(): Malformed hexstring: %s (length: %d)\n", hex, len);
+	cli_errmsg("cli_hex2si(): Malformed hexstring: %s (length: %d)\n", hex, len);
 	return NULL;
     }
 
@@ -68,12 +68,14 @@ short int *cl_hex2str(const char *hex)
     for(i = 0; i < len; i += 2) {
 	if(hex[i] == '?') {
 	    val = CLI_IGN;
+	} else if(hex[i] == '@') {
+	    val = CLI_ALT;
 	} else {
 	    if((c = cli_hex2int(hex[i])) >= 0) {
 		val = c;
 		if((c = cli_hex2int(hex[i+1])) >= 0) {
 		    val = (val << 4) + c;
-		} else { 
+		} else {
 		    free(str);
 		    return NULL;
 		}
@@ -88,7 +90,46 @@ short int *cl_hex2str(const char *hex)
     return str;
 }
 
-char *cl_str2hex(const char *string, unsigned int len)
+char *cli_hex2str(const char *hex)
+{
+	char *str, *ptr, val, c;
+	int i, len;
+
+
+    len = strlen(hex);
+
+    if(len % 2 != 0) {
+	cli_errmsg("cli_hex2str(): Malformed hexstring: %s (length: %d)\n", hex, len);
+	return NULL;
+    }
+
+    str = cli_calloc((len / 2) + 1, sizeof(char));
+    if(!str)
+	return NULL;
+
+    ptr = str;
+
+    for(i = 0; i < len; i += 2) {
+	if((c = cli_hex2int(hex[i])) >= 0) {
+	    val = c;
+	    if((c = cli_hex2int(hex[i+1])) >= 0) {
+		val = (val << 4) + c;
+	    } else {
+		free(str);
+		return NULL;
+	    }
+	} else {
+	    free(str);
+	    return NULL;
+	}
+
+	*ptr++ = val;
+    }
+
+    return str;
+}
+
+char *cli_str2hex(const char *string, unsigned int len)
 {
 	char *hexstr;
 	char HEX[] = { '0', '1', '2', '3', '4', '5', '6', '7', '8', '9',
@@ -122,27 +163,31 @@ int cli_strbcasestr(const char *haystack, const char *needle)
     return !strcasecmp(pt, needle);
 }
 
-void cli_chomp(char *string)
+/*
+ * Remove trailing NL and CR characters from the end of the given string.
+ * Return the new length of the string (ala strlen)
+ */
+int
+cli_chomp(char *string)
 {
-	size_t l = strlen(string);
+	int l;
 
+	if(string == NULL)
+		return -1;
 
-    if(l == 0)
-	return;
+	l  = strlen(string);
 
-    --l;
-    if((string[l] == '\n') || (string[l] == '\r')) {
-	string[l] = '\0';
+	if(l == 0)
+		return 0;
 
-	if(l > 0) {
-	    --l;
-	    if(string[l] == '\r')
-		string[l] = '\0';
-	}
-    }
+	--l;
+
+	while((l >= 0) && ((string[l] == '\n') || (string[l] == '\r')))
+		string[l--] = '\0';
+
+	return l + 1;
 }
 
-
 /*
  * char *cli_strok(const char *line, int fieldno, char *delim)
  * Return a copy of field <fieldno> from the string <line>, where
@@ -179,10 +224,85 @@ char *cli_strtok(const char *line, int fieldno, const char *delim)
 	return NULL;
     }
     buffer = malloc(j-i+1);
+    if(!buffer)
+	return NULL;
     strncpy(buffer, line+i, j-i);
     buffer[j-i] = '\0';
 
     return buffer;
 }
 
+/*
+ * Like cli_strtok, but this puts the output into a given argument, rather
+ * than allocating fresh memory
+ * Returns NULL for error, or a pointer to output
+ * njh@bandsman.co.uk
+ */
+char *cli_strtokbuf(const char *input, int fieldno, const char *delim, char *output)
+{
+    int counter = 0, i, j;
 
+    /* step to arg # <fieldno> */
+    for (i=0; input[i] && counter != fieldno; i++) {
+	if (strchr(delim, input[i])) {
+	    counter++;
+	    while(input[i+1] && strchr(delim, input[i+1])) {
+		i++;
+	    }
+	}
+    }
+    if (input[i] == '\0') {
+	/* end of buffer before field reached */
+	return NULL;
+    }
+
+    for (j=i; input[j]; j++) {
+	if (strchr(delim, input[j])) {
+	    break;
+	}
+    }
+    if (i == j) {
+	return NULL;
+    }
+    strncpy(output, input+i, j-i);
+    output[j-i] = '\0';
+
+    return output;
+}
+
+const char *cli_memstr(const char *haystack, int hs, const char *needle, int ns)
+{
+	const char *pt, *hay;
+	int n;
+
+
+    if(hs < ns)
+	return NULL;
+
+    if(haystack == needle)
+	return haystack;
+
+    if(!memcmp(haystack, needle, ns))
+	return haystack;
+
+    pt = hay = haystack;
+    n = hs;
+
+    while((pt = memchr(hay, needle[0], n)) != NULL) {
+	n -= (int) (pt - hay);
+	if(n < ns)
+	    break;
+
+	if(!memcmp(pt, needle, ns))
+	    return pt;
+
+	if(hay == pt) {
+	    n--;
+	    hay++;
+	} else {
+	    hay = pt;
+	}
+    }
+
+    return NULL;
+}
