@@ -16,6 +16,27 @@
  *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
  *
  * $Log: text.c,v $
+ * Revision 1.14  2005/01/19 05:31:55  nigelhorne
+ * Added textIterate
+ *
+ * Revision 1.13  2004/12/08 19:03:41  nigelhorne
+ * Fix compilation error on Solaris
+ *
+ * Revision 1.12  2004/12/04 16:03:55  nigelhorne
+ * Text/plain now handled as no encoding
+ *
+ * Revision 1.11  2004/11/27 21:54:26  nigelhorne
+ * Tidy
+ *
+ * Revision 1.10  2004/08/22 10:34:24  nigelhorne
+ * Use fileblob
+ *
+ * Revision 1.9  2004/08/21 11:57:57  nigelhorne
+ * Use line.[ch]
+ *
+ * Revision 1.8  2004/07/20 14:35:29  nigelhorne
+ * Some MYDOOM.I were getting through
+ *
  * Revision 1.7  2004/06/22 04:08:02  nigelhorne
  * Optimise empty lines
  *
@@ -30,37 +51,51 @@
  *
  */
 
-static	char	const	rcsid[] = "$Id: text.c,v 1.7 2004/06/22 04:08:02 nigelhorne Exp $";
+static	char	const	rcsid[] = "$Id: text.c,v 1.14 2005/01/19 05:31:55 nigelhorne Exp $";
 
 #if HAVE_CONFIG_H
 #include "clamav-config.h"
 #endif
 
+#ifndef	CL_DEBUG
+#define	NDEBUG	/* map CLAMAV debug onto standard */
+#endif
+
 #include <stdlib.h>
-#if	C_DARWIN
+#ifdef	C_DARWIN
 #include <sys/types.h>
 #include <sys/malloc.h>
 #else
 #ifdef HAVE_MALLOC_H /* tk: FreeBSD-CURRENT doesn't support malloc.h */
+#ifndef	C_BSD	/* BSD now uses stdlib.h */
 #include <malloc.h>
 #endif
 #endif
+#endif
 #include <string.h>
 #include <ctype.h>
 #include <assert.h>
+#include <stdio.h>
 
+#include "line.h"
 #include "mbox.h"
 #include "blob.h"
 #include "text.h"
 #include "others.h"
 
+static	text	*textCopy(const text *t_head);
+static	void	addToFileblob(const line_t *line, void *arg);
+static	void	getLength(const line_t *line, void *arg);
+static	void	addToBlob(const line_t *line, void *arg);
+static	void	*textIterate(const text *t_text, void (*cb)(const line_t *line, void *arg), void *arg);
+
 void
 textDestroy(text *t_head)
 {
 	while(t_head) {
 		text *t_next = t_head->t_next;
-		if(t_head->t_text)
-			free(t_head->t_text);
+		if(t_head->t_line)
+			lineUnlink(t_head->t_line);
 		free(t_head);
 		t_head = t_next;
 	}
@@ -68,75 +103,18 @@ textDestroy(text *t_head)
 
 /*
  * Remove trailing spaces from the lines and trailing blank lines
+ * This could be used to remove trailing blank lines, empty lines etc.,
+ *	but it probably isn't worth the time taken given that it won't reclaim
+ *	much memory
  */
 text *
 textClean(text *t_head)
 {
-#if	0	/* not needed in ClamAV */
-	text *t_lastnonempty = NULL, *t_ret;
-
-	while(t_head) {
-		char *line = t_head->t_text;
-		const size_t len = strlen(line);
-		int uuencoded = 0;
-
-		if((len > 0) && !uuencoded) {
-			int last = len;
-
-			if((strncasecmp(line, "begin ", 6) == 0) &&
-			   (isdigit(line[6])) &&
-			   (isdigit(line[7])) &&
-			   (isdigit(line[8])) &&
-			   (line[9] == ' '))
-				uuencoded = 1;
-			else
-				/*
-				 * Don't remove trailing spaces since that may
-				 * break uuencoded files
-				 */
-				while((--last >= 0) && isspace(line[last]))
-					;
-			if(++last > 0) {
-				t_lastnonempty = t_head;
-				if(last < len) {
-					line[last] = '\0';
-					t_head->t_text = cli_realloc(line, ++last);
-				}
-			} else {
-				t_head->t_text = cli_realloc(line, 1);
-				t_head->t_text[0] = '\0';
-			}
-		}
-		t_head = t_head->t_next;
-	}
-
-	if(t_lastnonempty == NULL)
-		return(NULL);	/* empty message I presume */
-
-	t_ret = t_lastnonempty;
-	t_lastnonempty = t_lastnonempty->t_next;
-
-	while(t_lastnonempty) {
-		text *t_next = t_lastnonempty->t_next;
-
-		assert(strlen(t_lastnonempty->t_text) == 0);
-
-		free(t_lastnonempty->t_text);
-		free(t_lastnonempty);
-
-		t_lastnonempty = t_next;
-	}
-
-	t_ret->t_next = NULL;
-
-	return t_ret;
-#else
 	return t_head;
-#endif
 }
 
 /* Clone the current object */
-text *
+static text *
 textCopy(const text *t_head)
 {
 	text *first = NULL, *last = NULL;
@@ -151,11 +129,10 @@ textCopy(const text *t_head)
 
 		assert(last != NULL);
 
-		if(t_head->t_text) {
-			last->t_text = strdup(t_head->t_text);
-			assert(last->t_text != NULL);
-		} else
-			last->t_text = NULL;
+		if(t_head->t_line)
+			last->t_line = lineLink(t_head->t_line);
+		else
+			last->t_line = NULL;
 
 		t_head = t_head->t_next;
 	}
@@ -171,6 +148,7 @@ text *
 textAdd(text *t_head, const text *t)
 {
 	text *ret;
+	int count;
 
 	if(t_head == NULL)
 		return textCopy(t);
@@ -180,8 +158,13 @@ textAdd(text *t_head, const text *t)
 
 	ret = t_head;
 
-	while(t_head->t_next)
+	count = 0;
+	while(t_head->t_next) {
+		count++;
 		t_head = t_head->t_next;
+	}
+
+	cli_dbgmsg("textAdd: count = %d\n", count);
 
 	while(t) {
 		t_head->t_next = (text *)cli_malloc(sizeof(text));
@@ -189,11 +172,10 @@ textAdd(text *t_head, const text *t)
 
 		assert(t_head != NULL);
 
-		if(t->t_text) {
-			t_head->t_text = strdup(t->t_text);
-			assert(t_head->t_text != NULL);
-		} else
-			t_head->t_text = NULL;
+		if(t->t_line)
+			t_head->t_line = lineLink(t->t_line);
+		else
+			t_head->t_line = NULL;
 
 		t = t->t_next;
 	}
@@ -207,7 +189,7 @@ textAdd(text *t_head, const text *t)
  * Add a message's content to the end of the current object
  */
 text *
-textAddMessage(text *aText, const message *aMessage)
+textAddMessage(text *aText, message *aMessage)
 {
 	assert(aMessage != NULL);
 
@@ -232,10 +214,17 @@ textAddMessage(text *aText, const message *aMessage)
 blob *
 textToBlob(const text *t, blob *b)
 {
-	const text *t1;
-	size_t s = 0;
+	size_t s;
 
-	assert(t != NULL);
+	if(t == NULL)
+		return NULL;
+
+	s = 0;
+
+	textIterate(t, getLength, &s);
+
+	if(s == 0)
+		return b;
 
 	if(b == NULL) {
 		b = blobCreate();
@@ -244,21 +233,74 @@ textToBlob(const text *t, blob *b)
 			return NULL;
 	}
 
-	for(t1 = t; t1; t1 = t1->t_next)
-		if(t1->t_text)
-			s += strlen(t1->t_text) + 1;
-		else
-			s++;
-
 	blobGrow(b, s);
 
-	do {
-		if(t->t_text)
-			blobAddData(b, (unsigned char *)t->t_text, strlen(t->t_text));
-		blobAddData(b, (unsigned char *)"\n", 1);
-	} while((t = t->t_next) != NULL);
+	textIterate(t, addToBlob, b);
 
 	blobClose(b);
 
 	return b;
 }
+
+fileblob *
+textToFileblob(const text *t, fileblob *fb)
+{
+	assert(fb != NULL);
+	assert(t != NULL);
+
+	if(fb == NULL) {
+		fb = fileblobCreate();
+
+		if(fb == NULL)
+			return NULL;
+	}
+
+	return textIterate(t, addToFileblob, fb);
+}
+
+static void
+getLength(const line_t *line, void *arg)
+{
+	size_t *length = (size_t *)arg;
+
+	if(line)
+		*length += strlen(lineGetData(line)) + 1;
+	else
+		(*length)++;
+}
+
+static void
+addToBlob(const line_t *line, void *arg)
+{
+	blob *b = (blob *)arg;
+
+	if(line) {
+		const char *l = lineGetData(line);
+
+		blobAddData(b, (unsigned char *)l, strlen(l));
+	}
+	blobAddData(b, (unsigned char *)"\n", 1);
+}
+
+static void
+addToFileblob(const line_t *line, void *arg)
+{
+	fileblob *fb = (fileblob *)arg;
+
+	if(line) {
+		const char *l = lineGetData(line);
+
+		fileblobAddData(fb, (unsigned char *)l, strlen(l));
+	}
+	fileblobAddData(fb, (unsigned char *)"\n", 1);
+}
+
+static void *
+textIterate(const text *t_text, void (*cb)(const line_t *item, void *arg), void *arg)
+{
+	while(t_text) {
+		(*cb)(t_text->t_line, arg);
+		t_text = t_text->t_next;
+	}
+	return arg;
+}
